@@ -45,12 +45,11 @@ use Errno ();
 
 use AnyEvent 5.0 ();
 use AnyEvent::Util ();
-use AnyEvent::Socket ();
 use AnyEvent::Handle ();
 
 use base Exporter::;
 
-our $VERSION = '1.45';
+our $VERSION = '1.5';
 
 our @EXPORT = qw(http_get http_post http_head http_request);
 
@@ -94,7 +93,7 @@ must be an absolute http or https URL.
 When called in void context, nothing is returned. In other contexts,
 C<http_request> returns a "cancellation guard" - you have to keep the
 object at least alive until the callback get called. If the object gets
-destroyed before the callbakc is called, the request will be cancelled.
+destroyed before the callback is called, the request will be cancelled.
 
 The callback will be called with the response body data as first argument
 (or C<undef> if an error occured), and a hash-ref with response headers as
@@ -103,7 +102,10 @@ second argument.
 All the headers in that hash are lowercased. In addition to the response
 headers, the "pseudo-headers" (uppercase to avoid clashing with possible
 response headers) C<HTTPVersion>, C<Status> and C<Reason> contain the
-three parts of the HTTP Status-Line of the same name.
+three parts of the HTTP Status-Line of the same name. If an error occurs
+during the body phase of a request, then the original C<Status> and
+C<Reason> values from the header are available as C<OrigStatus> and
+C<OrigReason>.
 
 The pseudo-header C<URL> contains the actual URL (which can differ from
 the requested URL when following redirects - for example, you might get
@@ -214,6 +216,18 @@ and behaves exactly the same way (e.g. it has to provide a
 timeout). See the description for the C<$prepare_cb> argument of
 C<AnyEvent::Socket::tcp_connect> for details.
 
+=item tcp_connect => $callback->($host, $service, $connect_cb, $prepare_cb)
+
+In even rarer cases you want total control over how AnyEvent::HTTP
+establishes connections. Normally it uses L<AnyEvent::Socket::tcp_connect>
+to do this, but you can provide your own C<tcp_connect> function -
+obviously, it has to follow the same calling conventions, except that it
+may always return a connection guard object.
+
+There are probably lots of weird uses for this function, starting from
+tracing the hosts C<http_request> actually tries to connect, to (inexact
+but fast) host => IP address caching or even socks protocol support.
+
 =item on_header => $callback->($headers)
 
 When specified, this callback will be called with the header hash as soon
@@ -340,7 +354,6 @@ sub _get_slot($$) {
    _slot_schedule $_[0];
 }
 
-our $qr_nl   = qr{\015?\012};
 our $qr_nlnl = qr{(?<![^\012])\015?\012};
 
 our $TLS_CTX_LOW  = { cache => 1, sslv2 => 1 };
@@ -369,23 +382,23 @@ sub http_request($$@) {
 
    my $recurse = exists $arg{recurse} ? delete $arg{recurse} : $MAX_RECURSE;
 
-   return $cb->(undef, { Status => 599, Reason => "Too many redirections", @pseudo })
+   return $cb->(undef, { @pseudo, Status => 599, Reason => "Too many redirections" })
       if $recurse < 0;
 
    my $proxy   = $arg{proxy}   || $PROXY;
    my $timeout = $arg{timeout} || $TIMEOUT;
 
    my ($uscheme, $uauthority, $upath, $query, $fragment) =
-      $url =~ m|(?:([^:/?#]+):)?(?://([^/?#]*))?([^?#]*)(?:\?([^#]*))?(?:#(.*))?|;
+      $url =~ m|(?:([^:/?#]+):)?(?://([^/?#]*))?([^?#]*)(?:(\?[^#]*))?(?:#(.*))?|;
 
    $uscheme = lc $uscheme;
 
    my $uport = $uscheme eq "http"  ?  80
              : $uscheme eq "https" ? 443
-             : return $cb->(undef, { Status => 599, Reason => "Only http and https URL schemes supported", @pseudo });
+             : return $cb->(undef, { @pseudo, Status => 599, Reason => "Only http and https URL schemes supported" });
 
    $uauthority =~ /^(?: .*\@ )? ([^\@:]+) (?: : (\d+) )?$/x
-      or return $cb->(undef, { Status => 599, Reason => "Unparsable URL", @pseudo });
+      or return $cb->(undef, { @pseudo, Status => 599, Reason => "Unparsable URL" });
 
    my $uhost = $1;
    $uport = $2 if defined $2;
@@ -394,7 +407,7 @@ sub http_request($$@) {
       unless exists $hdr{host};
 
    $uhost =~ s/^\[(.*)\]$/$1/;
-   $upath .= "?$query" if length $query;
+   $upath .= $query if length $query;
 
    $upath =~ s%^/?%/%;
 
@@ -457,12 +470,12 @@ sub http_request($$@) {
 
       return unless $state{connect_guard};
 
-      $state{connect_guard} = AnyEvent::Socket::tcp_connect $rhost, $rport, sub {
+      my $connect_cb = sub {
          $state{fh} = shift
             or do {
                my $err = "$!";
                %state = ();
-               return $cb->(undef, { Status => 599, Reason => $err, @pseudo });
+               return $cb->(undef, { @pseudo, Status => 599, Reason => $err });
             };
 
          pop; # free memory, save a tree
@@ -472,31 +485,31 @@ sub http_request($$@) {
          # get handle
          $state{handle} = new AnyEvent::Handle
             fh       => $state{fh},
-            timeout  => $timeout,
             peername => $rhost,
-            tls_ctx  => $arg{tls_ctx};
+            tls_ctx  => $arg{tls_ctx},
+            # these need to be reconfigured on keepalive handles
+            timeout  => $timeout,
+            on_error => sub {
+               %state = ();
+               $cb->(undef, { @pseudo, Status => 599, Reason => $_[2] });
+            },
+            on_eof   => sub {
+               %state = ();
+               $cb->(undef, { @pseudo, Status => 599, Reason => "Unexpected end-of-file" });
+            },
+         ;
 
          # limit the number of persistent connections
          # keepalive not yet supported
-         if ($KA_COUNT{$_[1]} < $MAX_PERSISTENT_PER_HOST) {
-            ++$KA_COUNT{$_[1]};
-            $state{handle}{ka_count_guard} = AnyEvent::Util::guard {
-               --$KA_COUNT{$_[1]}
-            };
-            $hdr{connection} = "keep-alive";
-         } else {
+#         if ($KA_COUNT{$_[1]} < $MAX_PERSISTENT_PER_HOST) {
+#            ++$KA_COUNT{$_[1]};
+#            $state{handle}{ka_count_guard} = AnyEvent::Util::guard {
+#               --$KA_COUNT{$_[1]}
+#            };
+#            $hdr{connection} = "keep-alive";
+#         } else {
             delete $hdr{connection};
-         }
-
-         # (re-)configure handle
-         $state{handle}->on_error (sub {
-            %state = ();
-            $cb->(undef, { Status => 599, Reason => $_[2], @pseudo });
-         });
-         $state{handle}->on_eof (sub {
-            %state = ();
-            $cb->(undef, { Status => 599, Reason => "Unexpected end-of-file", @pseudo });
-         });
+#         }
 
          $state{handle}->starttls ("connect") if $rscheme eq "https";
 
@@ -517,211 +530,215 @@ sub http_request($$@) {
 
             %hdr = (); # reduce memory usage, save a kitten, also make it possible to re-use
 
-            # status line
-            $state{handle}->push_read (line => $qr_nl, sub {
-               $_[1] =~ /^HTTP\/([0-9\.]+) \s+ ([0-9]{3}) (?: \s+ ([^\015\012]*) )?/ix
-                  or return (%state = (), $cb->(undef, { Status => 599, Reason => "Invalid server response ($_[1])", @pseudo }));
+            # status line and headers
+            $state{handle}->push_read (line => $qr_nlnl, sub {
+               my $keepalive = pop;
 
-               push @pseudo,
-                  HTTPVersion => $1,
-                  Status      => $2,
-                  Reason      => $3,
-               ;
+               for ("$_[1]") {
+                  y/\015//d; # weed out any \015, as they show up in the weirdest of places.
 
-               # headers, could be optimized a bit
-               $state{handle}->unshift_read (line => $qr_nlnl, sub {
-                  for ("$_[1]") {
-                     y/\015//d; # weed out any \015, as they show up in the weirdest of places.
+                  /^HTTP\/([0-9\.]+) \s+ ([0-9]{3}) (?: \s+ ([^\015\012]*) )? \015?\012/igxc
+                     or return (%state = (), $cb->(undef, { @pseudo, Status => 599, Reason => "Invalid server response" }));
 
-                     # things seen, not parsed:
-                     # p3pP="NON CUR OTPi OUR NOR UNI"
+                  push @pseudo,
+                     HTTPVersion => $1,
+                     Status      => $2,
+                     Reason      => $3,
+                  ;
 
-                     $hdr{lc $1} .= ",$2"
-                        while /\G
-                              ([^:\000-\037]*):
-                              [\011\040]*
-                              ((?: [^\012]+ | \012[\011\040] )*)
-                              \012
-                           /gxc;
+                  # things seen, not parsed:
+                  # p3pP="NON CUR OTPi OUR NOR UNI"
 
-                     /\G$/
-                       or return (%state = (), $cb->(undef, { Status => 599, Reason => "Garbled response headers", @pseudo }));
+                  $hdr{lc $1} .= ",$2"
+                     while /\G
+                           ([^:\000-\037]*):
+                           [\011\040]*
+                           ((?: [^\012]+ | \012[\011\040] )*)
+                           \012
+                        /gxc;
+
+                  /\G$/
+                    or return (%state = (), $cb->(undef, { @pseudo, Status => 599, Reason => "Garbled response headers" }));
+               }
+
+               # remove the "," prefix we added to all headers above
+               substr $_, 0, 1, ""
+                  for values %hdr;
+
+               # patch in all pseudo headers
+               %hdr = (%hdr, @pseudo);
+
+               # redirect handling
+               # microsoft and other shitheads don't give a shit for following standards,
+               # try to support some common forms of broken Location headers.
+               if ($hdr{location} !~ /^(?: $ | [^:\/?\#]+ : )/x) {
+                  $hdr{location} =~ s/^\.\/+//;
+
+                  my $url = "$rscheme://$uhost:$uport";
+
+                  unless ($hdr{location} =~ s/^\///) {
+                     $url .= $upath;
+                     $url =~ s/\/[^\/]*$//;
                   }
 
-                  # remove the "," prefix we added to all headers above
-                  substr $_, 0, 1, ""
-                     for values %hdr;
+                  $hdr{location} = "$url/$hdr{location}";
+               }
 
-                  # patch in all pseudo headers
-                  %hdr = (%hdr, @pseudo);
+               my $redirect;
 
-                  # redirect handling
-                  # microsoft and other shitheads don't give a shit for following standards,
-                  # try to support some common forms of broken Location headers.
-                  if ($hdr{location} !~ /^(?: $ | [^:\/?\#]+ : )/x) {
-                     $hdr{location} =~ s/^\.\/+//;
+               if ($recurse) {
+                  my $status = $hdr{Status};
 
-                     my $url = "$rscheme://$uhost:$uport";
+                  # industry standard is to redirect POST as GET for
+                  # 301, 302 and 303, in contrast to http/1.0 and 1.1.
+                  # also, the UA should ask the user for 301 and 307 and POST,
+                  # industry standard seems to be to simply follow.
+                  # we go with the industry standard.
+                  if ($status == 301 or $status == 302 or $status == 303) {
+                     # HTTP/1.1 is unclear on how to mutate the method
+                     $method = "GET" unless $method eq "HEAD";
+                     $redirect = 1;
+                  } elsif ($status == 307) {
+                     $redirect = 1;
+                  }
+               }
 
-                     unless ($hdr{location} =~ s/^\///) {
-                        $url .= $upath;
-                        $url =~ s/\/[^\/]*$//;
-                     }
+               my $finish = sub { # ($data, $err_status, $err_reason[, $keepalive])
+                  $state{handle}->destroy if $state{handle};
+                  %state = ();
 
-                     $hdr{location} = "$url/$hdr{location}";
+                  if (defined $_[1]) {
+                     $hdr{OrigStatus} = $hdr{Status}; $hdr{Status} = $_[1];
+                     $hdr{OrigReason} = $hdr{Reason}; $hdr{Reason} = $_[2];
                   }
 
-                  my $redirect;
+                  # set-cookie processing
+                  if ($arg{cookie_jar}) {
+                     for ($hdr{"set-cookie"}) {
+                        # parse NAME=VALUE
+                        my @kv;
 
-                  if ($recurse) {
-                     my $status = $hdr{Status};
+                        while (/\G\s* ([^=;,[:space:]]+) \s*=\s* (?: "((?:[^\\"]+|\\.)*)" | ([^=;,[:space:]]*) )/gcxs) {
+                           my $name = $1;
+                           my $value = $3;
 
-                     if (($status == 301 || $status == 302) && $method ne "POST") {
-                        # apparently, mozilla et al. just change POST to GET here
-                        # more research is needed before we do the same
-                        $redirect = 1;
-                     } elsif ($status == 303) {
-                        # even http/1.1 is unclear on how to mutate the method
-                        $method = "GET" unless $method eq "HEAD";
-                        $redirect = 1;
-                     } elsif ($status == 307 && $method =~ /^(?:GET|HEAD)$/) {
-                        $redirect = 1;
-                     }
-                  }
-
-                  my $finish = sub {
-                     $state{handle}->destroy if $state{handle};
-                     %state = ();
-
-                     # set-cookie processing
-                     if ($arg{cookie_jar}) {
-                        for ($_[1]{"set-cookie"}) {
-                           # parse NAME=VALUE
-                           my @kv;
-
-                           while (/\G\s* ([^=;,[:space:]]+) \s*=\s* (?: "((?:[^\\"]+|\\.)*)" | ([^=;,[:space:]]*) )/gcxs) {
-                              my $name = $1;
-                              my $value = $3;
-
-                              unless ($value) {
-                                 $value = $2;
-                                 $value =~ s/\\(.)/$1/gs;
-                              }
-
-                              push @kv, $name => $value;
-
-                              last unless /\G\s*;/gc;
+                           unless ($value) {
+                              $value = $2;
+                              $value =~ s/\\(.)/$1/gs;
                            }
 
-                           last unless @kv;
+                           push @kv, $name => $value;
 
-                           my $name = shift @kv;
-                           my %kv = (value => shift @kv, @kv);
-
-                           my $cdom;
-                           my $cpath = (delete $kv{path}) || "/";
-
-                           if (exists $kv{domain}) {
-                              $cdom = delete $kv{domain};
-       
-                              $cdom =~ s/^\.?/./; # make sure it starts with a "."
-
-                              next if $cdom =~ /\.$/;
-          
-                              # this is not rfc-like and not netscape-like. go figure.
-                              my $ndots = $cdom =~ y/.//;
-                              next if $ndots < ($cdom =~ /\.[^.][^.]\.[^.][^.]$/ ? 3 : 2);
-                           } else {
-                              $cdom = $uhost;
-                           }
-       
-                           # store it
-                           $arg{cookie_jar}{version} = 1;
-                           $arg{cookie_jar}{$cdom}{$cpath}{$name} = \%kv;
-
-                           redo if /\G\s*,/gc;
+                           last unless /\G\s*;/gc;
                         }
+
+                        last unless @kv;
+
+                        my $name = shift @kv;
+                        my %kv = (value => shift @kv, @kv);
+
+                        my $cdom;
+                        my $cpath = (delete $kv{path}) || "/";
+
+                        if (exists $kv{domain}) {
+                           $cdom = delete $kv{domain};
+    
+                           $cdom =~ s/^\.?/./; # make sure it starts with a "."
+
+                           next if $cdom =~ /\.$/;
+       
+                           # this is not rfc-like and not netscape-like. go figure.
+                           my $ndots = $cdom =~ y/.//;
+                           next if $ndots < ($cdom =~ /\.[^.][^.]\.[^.][^.]$/ ? 3 : 2);
+                        } else {
+                           $cdom = $uhost;
+                        }
+    
+                        # store it
+                        $arg{cookie_jar}{version} = 1;
+                        $arg{cookie_jar}{$cdom}{$cpath}{$name} = \%kv;
+
+                        redo if /\G\s*,/gc;
                      }
+                  }
 
-                     if ($redirect && exists $hdr{location}) {
-                        # we ignore any errors, as it is very common to receive
-                        # Content-Length != 0 but no actual body
-                        # we also access %hdr, as $_[1] might be an erro
-                        http_request (
-                           $method  => $hdr{location},
-                           %arg,
-                           recurse  => $recurse - 1,
-                           Redirect => \@_,
-                           $cb);
-                     } else {
-                        $cb->($_[0], $_[1]);
-                     }
-                  };
-
-                  my $len = $hdr{"content-length"};
-
-                  if (!$redirect && $arg{on_header} && !$arg{on_header}(\%hdr)) {
-                     $finish->(undef, { Status => 598, Reason => "Request cancelled by on_header", @pseudo });
-                  } elsif (
-                     $hdr{Status} =~ /^(?:1..|[23]04)$/
-                     or $method eq "HEAD"
-                     or (defined $len && !$len)
-                  ) {
-                     # no body
-                     $finish->("", \%hdr);
+                  if ($redirect && exists $hdr{location}) {
+                     # we ignore any errors, as it is very common to receive
+                     # Content-Length != 0 but no actual body
+                     # we also access %hdr, as $_[1] might be an erro
+                     http_request (
+                        $method  => $hdr{location},
+                        %arg,
+                        recurse  => $recurse - 1,
+                        Redirect => [$_[0], \%hdr],
+                        $cb);
                   } else {
-                     # body handling, four different code paths
-                     # for want_body_handle, on_body (2x), normal (2x)
-                     # we might read too much here, but it does not matter yet (no pers. connections)
-                     if (!$redirect && $arg{want_body_handle}) {
-                        $_[0]->on_eof (undef);
-                        $_[0]->on_error (undef);
-                        $_[0]->on_read  (undef);
+                     $cb->($_[0], \%hdr);
+                  }
+               };
 
-                        $finish->(delete $state{handle}, \%hdr);
+               my $len = $hdr{"content-length"};
 
-                     } elsif ($arg{on_body}) {
-                        $_[0]->on_error (sub { $finish->(undef, { Status => 599, Reason => $_[2], @pseudo }) });
-                        if ($len) {
-                           $_[0]->on_eof (undef);
-                           $_[0]->on_read (sub {
-                              $len -= length $_[0]{rbuf};
+               if (!$redirect && $arg{on_header} && !$arg{on_header}(\%hdr)) {
+                  $finish->(undef, 598 => "Request cancelled by on_header");
+               } elsif (
+                  $hdr{Status} =~ /^(?:1..|204|205|304)$/
+                  or $method eq "HEAD"
+                  or (defined $len && !$len)
+               ) {
+                  # no body
+                  $finish->("", undef, undef, 1);
+               } else {
+                  # body handling, four different code paths
+                  # for want_body_handle, on_body (2x), normal (2x)
+                  # we might read too much here, but it does not matter yet (no pipelining)
+                  if (!$redirect && $arg{want_body_handle}) {
+                     $_[0]->on_eof   (undef);
+                     $_[0]->on_error (undef);
+                     $_[0]->on_read  (undef);
 
-                              $arg{on_body}(delete $_[0]{rbuf}, \%hdr)
-                                 or $finish->(undef, { Status => 598, Reason => "Request cancelled by on_body", @pseudo });
+                     $finish->(delete $state{handle});
 
-                              $len > 0
-                                 or $finish->("", \%hdr);
-                           });
-                        } else {
-                           $_[0]->on_eof (sub {
-                              $finish->("", \%hdr);
-                           });
-                           $_[0]->on_read (sub {
-                              $arg{on_body}(delete $_[0]{rbuf}, \%hdr)
-                                 or $finish->(undef, { Status => 598, Reason => "Request cancelled by on_body", @pseudo });
-                           });
-                        }
+                  } elsif ($arg{on_body}) {
+                     $_[0]->on_error (sub { $finish->(undef, 599 => $_[2]) });
+                     if ($len) {
+                        $_[0]->on_read (sub {
+                           $len -= length $_[0]{rbuf};
+
+                           $arg{on_body}(delete $_[0]{rbuf}, \%hdr)
+                              or $finish->(undef, 598 => "Request cancelled by on_body");
+
+                           $len > 0
+                              or $finish->("", undef, undef, 1);
+                        });
                      } else {
-                        $_[0]->on_eof (undef);
+                        $_[0]->on_eof (sub {
+                           $finish->("");
+                        });
+                        $_[0]->on_read (sub {
+                           $arg{on_body}(delete $_[0]{rbuf}, \%hdr)
+                              or $finish->(undef, 598 => "Request cancelled by on_body");
+                        });
+                     }
+                  } else {
+                     $_[0]->on_eof (undef);
 
-                        if ($len) {
-                           $_[0]->on_error (sub { $finish->(undef, { Status => 599, Reason => $_[2], @pseudo }) });
-                           $_[0]->on_read (sub {
-                              $finish->((substr delete $_[0]{rbuf}, 0, $len, ""), \%hdr)
-                                 if $len <= length $_[0]{rbuf};
-                           });
-                        } else {
-                           $_[0]->on_error (sub {
-                              $! == Errno::EPIPE || !$!
-                                 ? $finish->(delete $_[0]{rbuf}, \%hdr)
-                                 : $finish->(undef, { Status => 599, Reason => $_[2], @pseudo });
-                           });
-                           $_[0]->on_read (sub { });
-                        }
+                     if ($len) {
+                        $_[0]->on_error (sub { $finish->(undef, 599 => $_[2]) });
+                        $_[0]->on_read (sub {
+                           $finish->((substr delete $_[0]{rbuf}, 0, $len, ""), undef, undef, 1)
+                              if $len <= length $_[0]{rbuf};
+                        });
+                     } else {
+                        $_[0]->on_error (sub {
+                           ($! == Errno::EPIPE || !$!)
+                              ? $finish->(delete $_[0]{rbuf})
+                              : $finish->(undef, 599 => $_[2]);
+                        });
+                        $_[0]->on_read (sub { });
                      }
                   }
-               });
+               }
             });
          };
 
@@ -733,21 +750,26 @@ sub http_request($$@) {
             $state{handle}->push_write ("CONNECT $uhost:$uport HTTP/1.0\015\012Host: $uhost\015\012\015\012");
             $state{handle}->push_read (line => $qr_nlnl, sub {
                $_[1] =~ /^HTTP\/([0-9\.]+) \s+ ([0-9]{3}) (?: \s+ ([^\015\012]*) )?/ix
-                  or return (%state = (), $cb->(undef, { Status => 599, Reason => "Invalid proxy connect response ($_[1])", @pseudo }));
+                  or return (%state = (), $cb->(undef, { @pseudo, Status => 599, Reason => "Invalid proxy connect response ($_[1])" }));
 
                if ($2 == 200) {
                   $rpath = $upath;
                   &$handle_actual_request;
                } else {
                   %state = ();
-                  $cb->(undef, { Status => $2, Reason => $3, @pseudo });
+                  $cb->(undef, { @pseudo, Status => $2, Reason => $3 });
                }
             });
          } else {
             &$handle_actual_request;
          }
+      };
+
+      my $tcp_connect = $arg{tcp_connect}
+                        || do { require AnyEvent::Socket; \&AnyEvent::Socket::tcp_connect };
+
+      $state{connect_guard} = $tcp_connect->($rhost, $rport, $connect_cb, $arg{on_prepare} || sub { $timeout });
 
-      }, $arg{on_prepare} || sub { $timeout };
    };
 
    defined wantarray && AnyEvent::Util::guard { %state = () }
@@ -792,6 +814,16 @@ otherwise.
 
 To clear an already-set proxy, use C<undef>.
 
+=item $date = AnyEvent::HTTP::format_date $timestamp
+
+Takes a POSIX timestamp (seconds since the epoch) and formats it as a HTTP
+Date (RFC 2616).
+
+=item $timestamp = AnyEvent::HTTP::parse_date $date
+
+Takes a HTTP Date (RFC 2616) and returns the corresponding POSIX
+timestamp, or C<undef> if the date cannot be parsed.
+
 =item $AnyEvent::HTTP::MAX_RECURSE
 
 The default value for the C<recurse> request parameter (default: C<10>).
@@ -820,6 +852,49 @@ connections. This number of can be useful for load-leveling.
 
 =cut
 
+our @month   = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
+our @weekday = qw(Sun Mon Tue Wed Thu Fri Sat);
+
+sub format_date($) {
+   my ($time) = @_;
+
+   # RFC 822/1123 format
+   my ($S, $M, $H, $mday, $mon, $year, $wday, $yday, undef) = gmtime $time;
+
+   sprintf "%s, %02d %s %04d %02d:%02d:%02d GMT",
+      $weekday[$wday], $mday, $month[$mon], $year + 1900,
+      $H, $M, $S;
+}
+
+sub parse_date($) {
+   my ($date) = @_;
+
+   my ($d, $m, $y, $H, $M, $S);
+
+   if ($date =~ /^[A-Z][a-z][a-z], ([0-9][0-9]) ([A-Z][a-z][a-z]) ([0-9][0-9][0-9][0-9]) ([0-9][0-9]):([0-9][0-9]):([0-9][0-9]) GMT$/) {
+      # RFC 822/1123, required by RFC 2616
+      ($d, $m, $y, $H, $M, $S) = ($1, $2, $3, $4, $5, $6);
+
+   } elsif ($date =~ /^[A-Z][a-z]+, ([0-9][0-9])-([A-Z][a-z][a-z])-([0-9][0-9]) ([0-9][0-9]):([0-9][0-9]):([0-9][0-9]) GMT$/) {
+      # RFC 850
+      ($d, $m, $y, $H, $M, $S) = ($1, $2, $3 < 69 ? $3 + 2000 : $3 + 1900, $4, $5, $6);
+
+   } elsif ($date =~ /^[A-Z][a-z][a-z] ([A-Z][a-z][a-z]) ([0-9 ][0-9]) ([0-9][0-9]):([0-9][0-9]):([0-9][0-9]) ([0-9][0-9][0-9][0-9])$/) {
+      # ISO C's asctime
+      ($d, $m, $y, $H, $M, $S) = ($2, $1, $6, $3, $4, $5);
+   }
+   # other formats fail in the loop below
+
+   for (0..11) {
+      if ($m eq $month[$_]) {
+         require Time::Local;
+         return Time::Local::timegm ($S, $M, $H, $d, $_, $y);
+      }
+   }
+
+   undef
+}
+
 sub set_proxy($) {
    if (length $_[0]) {
       $_[0] =~ m%^(https?):// ([^:/]+) (?: : (\d*) )?%ix
@@ -835,6 +910,62 @@ eval {
    set_proxy $ENV{http_proxy};
 };
 
+=head2 SOCKS PROXIES
+
+Socks proxies are not directly supported by AnyEvent::HTTP. You can
+compile your perl to support socks, or use an external program such as
+F<socksify> (dante) or F<tsocks> to make your program use a socks proxy
+transparently.
+
+Alternatively, for AnyEvent::HTTP only, you can use your own
+C<tcp_connect> function that does the proxy handshake - here is an example
+that works with socks4a proxies:
+
+   use Errno;
+   use AnyEvent::Util;
+   use AnyEvent::Socket;
+   use AnyEvent::Handle;
+
+   # host, port and username of/for your socks4a proxy
+   my $socks_host = "10.0.0.23";
+   my $socks_port = 9050;
+   my $socks_user = "";
+
+   sub socks4a_connect {
+      my ($host, $port, $connect_cb, $prepare_cb) = @_;
+
+      my $hdl = new AnyEvent::Handle
+         connect    => [$socks_host, $socks_port],
+         on_prepare => sub { $prepare_cb->($_[0]{fh}) },
+         on_error   => sub { $connect_cb->() },
+      ;
+
+      $hdl->push_write (pack "CCnNZ*Z*", 4, 1, $port, 1, $socks_user, $host);
+
+      $hdl->push_read (chunk => 8, sub {
+         my ($hdl, $chunk) = @_;
+         my ($status, $port, $ipn) = unpack "xCna4", $chunk;
+
+         if ($status == 0x5a) {
+            $connect_cb->($hdl->{fh}, (format_address $ipn) . ":$port");
+         } else {
+            $! = Errno::ENXIO; $connect_cb->();
+         }
+      });
+
+      $hdl
+   }
+
+Use C<socks4a_connect> instead of C<tcp_connect> when doing C<http_request>s,
+possibly after switching off other proxy types:
+
+   AnyEvent::HTTP::set_proxy undef; # usually you do not want other proxies
+
+   http_get 'http://www.google.com', tcp_connect => \&socks4a_connect, sub {
+      my ($data, $headers) = @_;
+      ...
+   };
+
 =head1 SEE ALSO
 
 L<AnyEvent>.
